from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING

from jinja2.ext import Extension, InternationalizationExtension


try:
    from babel.core import Locale, UnknownLocaleError
    from babel.support import NullTranslations, Translations

    HAS_BABEL = True
except ImportError:  # pragma: no cover
    from string import ascii_letters
    from typing import NamedTuple

    HAS_BABEL = False

    class UnknownLocaleError(Exception):  # type: ignore[no-redef]
        pass

    class Locale(NamedTuple):  # type: ignore[no-redef]
        language: str
        territory: str = ""

        def __str__(self):
            if self.territory:
                return f"{self.language}_{self.territory}"
            return self.language

        @classmethod
        def parse(cls, identifier: str, sep: str | None):
            if not isinstance(identifier, str):
                msg = f"Unexpected value for identifier: {identifier!r}"
                raise TypeError(msg)
            locale = cls(*identifier.split(sep, 1))
            if not all(x in ascii_letters for x in locale.language):
                msg = f"expected only letters, got {locale.language!r}"
                raise ValueError(msg)
            if len(locale.language) != 2:  # noqa: PLR2004
                msg = f"unknown locale {locale.language!r}"
                raise UnknownLocaleError(msg)
            return locale


if TYPE_CHECKING:
    from collections.abc import Sequence
    import os

    import jinja2


log = logging.getLogger(__name__)
BASE_PATH = pathlib.Path(__file__).absolute().parent


class NoBabelExtension(InternationalizationExtension):
    def __init__(self, environment: jinja2.Environment):
        Extension.__init__(self, environment)
        environment.extend(
            install_null_translations=self._install_null,
            newstyle_gettext=False,
        )


def parse_locale(locale: str) -> Locale:
    """Parse a locale string into a Locale object.

    Args:
        locale: The locale string to parse (e.g., 'en_US').

    Returns:
        Locale: The parsed Babel Locale object.

    Raises:
        RuntimeError: If the locale string is invalid.
    """
    try:
        return Locale.parse(locale, sep="_")
    except (ValueError, UnknownLocaleError, TypeError) as e:
        msg = f"Invalid value for locale: {e}"
        raise RuntimeError(msg) from e


def install_translations(
    env: jinja2.Environment, locale: str | Locale, dirs: Sequence[str | os.PathLike[str]]
) -> None:
    """Install translations for the given locale in the Jinja environment.

    Args:
        env: The Jinja2 environment to install translations into.
        locale: The target locale.
        dirs: Sequence of directory paths to search for translations.
              Directories listed first
    """
    if isinstance(locale, str):
        locale = parse_locale(locale)
    if HAS_BABEL:
        env.add_extension("jinja2.ext.i18n")
        translations = _get_merged_translations(dirs, "locales", locale)
        if translations is not None:
            env.install_gettext_translations(translations)  # type: ignore[attr-defined]
        else:
            env.install_null_translations()  # type: ignore[attr-defined]
            if locale.language != "en":
                log.warning(
                    "Translations couldnt be found for locale %r, defaulting to English",
                    locale,
                )
    else:  # pragma: no cover
        # no babel installed, add dummy support for trans/endtrans blocks
        env.add_extension(NoBabelExtension)
        env.install_null_translations()  # type: ignore[attr-defined]


def _get_merged_translations(
    dirs: Sequence[str | os.PathLike[str]], locales_dir: str, locale: Locale
) -> Translations | None:
    """Merge translations from multiple directories for a given locale.

    Args:
        dirs: Sequence of directory paths to search for translations.
        locales_dir: Name of the directory containing locale files.
        locale: The target locale.

    Returns:
        Merged translations object or None if no translations are found.
    """
    merged_translations: NullTranslations | None = None

    log.debug("Looking for translations for locale %r", locale)
    locale_str = (
        f"{locale.language}_{locale.territory}" if locale.territory else locale.language
    )
    for theme_dir in reversed(dirs):
        dirname = pathlib.Path(theme_dir) / locales_dir
        translations = Translations.load(dirname, [locale_str])

        if type(translations) is NullTranslations:
            log.debug("No translations found in: %r", dirname)
            continue
        log.debug("Translations found in: %r", dirname)
        if merged_translations is None:
            merged_translations = translations
        else:
            merged_translations.merge(translations)  # type: ignore[attr-defined]

    return merged_translations  # type: ignore[return-value]



from __future__ import annotations

import re
from typing import Any, Literal

from jinja2.exceptions import FilterArgumentError


def re_replace(
    value: str = "",
    pattern: str = "",
    replacement: str = "",
    *,
    ignorecase: bool = False,
    multiline: bool = False,
    count: int = 0,
) -> str:
    """Perform a `re.sub` returning a string.

    Filter adapted from Ansible

    Arguments:
        value: The value to search-replace.
        pattern: The regex pattern to use
        replacement: The replacement pattern to use
        ignorecase: Whether to ignore casing
        multiline: Whether to do a multiline regex search
        count: Amount of maximum substitutes.
    """
    flags = 0
    if ignorecase:
        flags |= re.I
    if multiline:
        flags |= re.M
    pat = re.compile(pattern, flags=flags)
    output, _subs = pat.subn(replacement, value, count=count)
    return output


def re_findall(
    value: str,
    regex: str,
    *,
    ignorecase: bool = False,
    multiline: bool = False,
) -> list[Any]:
    """Perform re.findall and return the list of matches.

    Filter adapted from Ansible

    Arguments:
        value: The text to search in
        regex: The regex to use for searching
        ignorecase: Whether char casing should be ignored
        multiline: Whether to perform a multi-line search
    """
    flags = 0
    if ignorecase:
        flags |= re.I
    if multiline:
        flags |= re.M
    return re.findall(regex, value, flags)


def re_search(
    value: str,
    regex: str,
    *args: str,
    ignorecase: bool = False,
    multiline: bool = False,
) -> list[str] | None | str:
    """Perform re.search and return the list of matches or a backref.

    Filter adapted from Ansible

    Arguments:
        value: The text to search in
        regex: The regex to use for searching
        args: Optional back references to return
        ignorecase: Whether char casing should be ignored
        multiline: Whether to perform a multi-line search
    """
    groups = list()
    for arg in args:
        if arg.startswith("\\g"):
            if match := re.match(r"\\g<(\S+)>", arg):
                groups.append(match.group(1))
        elif arg.startswith("\\"):
            if match := re.match(r"\\(\d+)", arg):
                groups.append(int(match.group(1)))
        else:
            msg = "Unknown argument"
            raise FilterArgumentError(msg)

    flags = 0
    if ignorecase:
        flags |= re.I
    if multiline:
        flags |= re.M
    if match := re.search(regex, value, flags):
        if not groups:
            return match.group()
        return [match.group(item) for item in groups]
    return None


def re_escape(string: str, re_type: Literal["python", "posix_basic"] = "python") -> str:
    """Escape all regular expressions special characters from STRING.

    Filter adapted from Ansible

    Arguments:
        string: The string to escape
        re_type: The escape type
    """
    match re_type:
        case "python":
            return re.escape(string)
        case "posix_basic":
            # list of BRE special chars:
            return re_replace(string, r"([].[^$*\\])", r"\\\1")
        case _:
            msg = f"Invalid regex type ({re_type})"
            raise NotImplementedError(msg)



from __future__ import annotations

import pytest

from jinjarope import envglobals


def test_match():
    assert envglobals.match("a", a="hit", b="miss") == "hit"


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import builtins
import datetime
import inspect
import math
import os
import pathlib
import re
import sys
from typing import TYPE_CHECKING, Any, TypeVar

from jinjarope import utils


T = TypeVar("T")


if TYPE_CHECKING:
    from collections.abc import Callable, Sequence


_RFC_3986_PATTERN = re.compile(r"^[A-Za-z][A-Za-z0-9+\-+.]*://")


def is_number(value: Any) -> bool:
    """Try to convert value to a float."""
    try:
        fvalue = float(value)
    except (ValueError, TypeError):
        return False
    return math.isfinite(fvalue)


def _is_type(value: Any) -> bool:
    """Return whether a value is a type."""
    return isinstance(value, type)


def _is_list(value: Any) -> bool:
    """Return whether a value is a list."""
    return isinstance(value, list)


def _is_set(value: Any) -> bool:
    """Return whether a value is a set."""
    return isinstance(value, set)


def _is_tuple(value: Any) -> bool:
    """Return whether a value is a tuple."""
    return isinstance(value, tuple)


def _is_dict(value: Any) -> bool:
    """Return whether a value is a tuple."""
    return isinstance(value, dict)


def _to_set(value: Any) -> set[Any]:
    """Convert value to set."""
    return set(value)


def _to_tuple(value: Sequence[T]) -> tuple[T, ...]:
    """Convert value to tuple."""
    return tuple(value)


def is_instance(obj: object, typ: str | type) -> bool:
    """Like the isinstance builtin, but also accepts strs as type.

    Arguments:
        obj: The object to check
        typ: A type (name)
    """
    kls = utils.resolve(typ) if isinstance(typ, str) else typ
    if not isinstance(kls, type):
        raise TypeError(kls)
    return isinstance(obj, kls)


def is_subclass(obj: type, typ: str | type) -> bool:
    """Like the issubclass builtin, but also accepts strs as type.

    Arguments:
        obj: The class to check
        typ: A type (name)
    """
    kls = utils.resolve(typ) if isinstance(typ, str) else typ
    if not isinstance(kls, type):
        raise TypeError(kls)
    return issubclass(obj, kls)


def _is_datetime(value: Any) -> bool:
    """Return whether a value is a datetime."""
    return isinstance(value, datetime.datetime)


def _is_string_like(value: Any) -> bool:
    """Return whether a value is a string or string like object."""
    return isinstance(value, str | bytes | bytearray)


def is_http_url(string: str) -> bool:
    """Return true when given string represents a HTTP url.

    Arguments:
        string: The string to check
    """
    return string.startswith(("http://", "https://", "www.")) and "\n" not in string


def is_protocol_url(string: str) -> bool:
    """Return true when given string represents any type of URL.

    Arguments:
        string: The string to check
    """
    return "://" in string and "\n" not in string


def is_python_keyword(string: str) -> bool:
    """Return true when given string represents a python keyword.

    Arguments:
        string: The string to check
    """
    import keyword

    return keyword.iskeyword(string)


def is_python_builtin(fn: str | Callable[..., Any]) -> bool:
    """Return true when given fn / string represents a python builtin.

    Arguments:
        fn: (Name of) function to check
    """
    return fn in dir(builtins) if isinstance(fn, str) else inspect.isbuiltin(fn)


def is_in_std_library(fn: str | Callable[..., Any]) -> bool:
    """Return true when given fn / string is part of the std library.

    Arguments:
        fn: (Name of) function to check
    """
    name = fn if isinstance(fn, str) else fn.__module__
    return name.split(".")[0] in sys.stdlib_module_names


def is_fsspec_url(string: str | os.PathLike[str]) -> bool:
    """Returns true if the given URL looks like an fsspec protocol, except http/https.

    Arguments:
        string: The URL to check
    """
    return (
        isinstance(string, str)
        and bool(_RFC_3986_PATTERN.match(string))
        and not string.startswith(("http://", "https://"))
    )


def contains_files(directory: str | os.PathLike[str]) -> bool:
    """Check if given directory exists and contains any files.

    Arguments:
        directory: The directoy to check
    """
    path = pathlib.Path(directory)
    return path.exists() and any(path.iterdir())


def is_installed(package_name: str) -> bool:
    """Returns true if a package with given name is found.

    Arguments:
        package_name: The package name to check
    """
    import importlib.util

    return bool(importlib.util.find_spec(package_name))


def is_env_var(env_var: str) -> bool:
    """Returns true if an environment variable with given name has a value.

    Arguments:
        env_var: The environment variable name to check
    """
    return bool(os.getenv(env_var))


def is_indented(text: str, indentation: str = "    ") -> bool:
    """Check whether all lines of given text are indented.

    Arguments:
        text: The text to check
        indentation: The indent each line must start with
    """
    return all(i.startswith(indentation) for i in text.split("\n"))


if __name__ == "__main__":
    result = is_in_std_library(inspect.getsource)
    print(result)



from __future__ import annotations

import mknodes as mk
import pytest

from jinjarope import manual


def test_building_the_docs():
    build = manual.Build()
    nav = mk.MkNav()
    build.on_root(nav)
    assert nav.children
    for node in nav.descendants:
        if isinstance(node, mk.MkPage):
            str(node)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from jinja2.exceptions import FilterArgumentError
import pytest

from jinjarope import regexfilters


def test_re_replace():
    assert (
        regexfilters.re_replace("Hello, World!", "World", "Universe")
        == "Hello, Universe!"
    )
    assert (
        regexfilters.re_replace("Hello, World!", "world", "Universe", ignorecase=True)
        == "Hello, Universe!"
    )
    assert (
        regexfilters.re_replace(
            "Hello, World! Hello, World!",
            "World",
            "Universe",
            count=1,
        )
        == "Hello, Universe! Hello, World!"
    )


def test_re_findall():
    assert regexfilters.re_findall("Hello, World! Hello, Universe!", "Hello") == [
        "Hello",
        "Hello",
    ]
    assert regexfilters.re_findall(
        "Hello, World! Hello, Universe!",
        "hello",
        ignorecase=True,
    ) == ["Hello", "Hello"]
    assert regexfilters.re_findall("Hello, World! Hello, Universe!", "Goodbye") == []


def test_re_search():
    assert regexfilters.re_search("Hello, World! Hello, Universe!", "World") == "World"
    assert (
        regexfilters.re_search("Hello, World! Hello, Universe!", "world", ignorecase=True)
        == "World"
    )
    assert regexfilters.re_search("Hello, World! Hello, Universe!", "Goodbye") is None
    assert regexfilters.re_search("A\nB", "^B", multiline=True)
    with pytest.raises(FilterArgumentError):
        regexfilters.re_search("A\nB", "^B", "x")


def test_re_escape():
    assert regexfilters.re_escape("Hello, World!") == "Hello,\\ World!"
    assert regexfilters.re_escape("Hello, World!", "posix_basic") == "Hello, World!"
    assert regexfilters.re_escape("[a]", "posix_basic") == "\\[a\\]"
    with pytest.raises(NotImplementedError):
        regexfilters.re_escape("Hello, World!", "unknown")


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import pytest

import jinjarope


def test_loading_filters():
    file = jinjarope.JinjaFile("src/jinjarope/resources/filters.toml")
    assert isinstance(file.filters[0], jinjarope.JinjaItem)
    assert file.filters_dict


def test_loading_functions():
    file = jinjarope.JinjaFile("src/jinjarope/resources/functions.toml")
    assert isinstance(file.functions[0], jinjarope.JinjaItem)
    assert file.functions_dict


def test_loading_tests():
    file = jinjarope.JinjaFile("src/jinjarope/resources/tests.toml")
    assert isinstance(file.tests[0], jinjarope.JinjaItem)
    assert file.tests_dict


def test_loading_config():
    file = jinjarope.JinjaFile("tests/testresources/testconfig.toml")
    assert file.envconfig.trim_blocks


def test_loading_loaders():
    env = jinjarope.Environment()
    env.load_jinja_file("tests/testresources/testconfig.toml")
    assert env.get_template("testfile.jinja").render()


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from xml.etree import ElementTree as ET

import pytest

from jinjarope import htmlfilters


def test_wrap_in_elem():
    assert htmlfilters.wrap_in_elem("Hello", "p") == "<p>Hello</p>"
    assert (
        htmlfilters.wrap_in_elem("Hello", "p", add_linebreaks=True) == "<p>\nHello\n</p>"
    )
    assert (
        htmlfilters.wrap_in_elem("Hello", "p", id="greeting", class_="greet")
        == '<p id="greeting" class="greet">Hello</p>'
    )
    assert htmlfilters.wrap_in_elem("", "p") == ""
    assert htmlfilters.wrap_in_elem(None, "p") == ""


def test_html_link():
    assert (
        htmlfilters.html_link("Google", "http://google.com")
        == "<a href='http://google.com'>Google</a>"
    )
    assert htmlfilters.html_link("Google", "") == "Google"
    assert htmlfilters.html_link("Google", None) == "Google"
    assert (
        htmlfilters.html_link(None, "http://google.com")
        == "<a href='http://google.com'>http://google.com</a>"
    )
    assert htmlfilters.html_link(None, None) == ""


def test_format_js_map():
    assert htmlfilters.format_js_map({"key": "value"}) == "{\n    key: 'value',\n}"
    assert htmlfilters.format_js_map('{"key": "value"}') == "{\n    key: 'value',\n}"
    assert htmlfilters.format_js_map({"key": True}) == "{\n    key: true,\n}"
    assert htmlfilters.format_js_map({"key": None}) == "{\n    key: null,\n}"
    assert (
        htmlfilters.format_js_map({"key": {"nested_key": "nested_value"}})
        == "{\n    key: {\n    nested_key: 'nested_value',\n},\n}"
    )


def test_svg_to_data_uri():
    assert (
        htmlfilters.svg_to_data_uri("<svg></svg>")
        == "url('data:image/svg+xml;charset=utf-8,<svg></svg>')"
    )
    assert htmlfilters.svg_to_data_uri("") == "url('data:image/svg+xml;charset=utf-8,')"
    with pytest.raises(TypeError):
        htmlfilters.svg_to_data_uri(None)  # type: ignore[arg-type]


def test_clean_svg():
    assert (
        htmlfilters.clean_svg(
            '<?xml version="1.0" encoding="UTF-8" standalone="no"?><svg></svg>',
        )
        == "<svg></svg>"
    )
    assert (
        htmlfilters.clean_svg(
            '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"'
            ' "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg></svg>',
        )
        == "<svg></svg>"
    )
    assert htmlfilters.clean_svg("<svg></svg>") == "<svg></svg>"
    assert htmlfilters.clean_svg("") == ""


def test_format_css_rule():
    assert htmlfilters.format_css_rule({".a": {"b": "c"}}) == ".a {\n\tb: c;\n}\n\n"
    assert (
        htmlfilters.format_css_rule({".a": {"b": "c", "d": "e"}})
        == ".a {\n\tb: c;\n\td: e;\n}\n\n"
    )
    assert (
        htmlfilters.format_css_rule({".a": {"b": {"c": "d"}}}) == ".a b {\n\tc: d;\n}\n\n"
    )
    assert htmlfilters.format_css_rule({}) == ""


def test_format_xml_with_str():
    xml_str = "<root><child>text</child></root>"
    formatted_xml = htmlfilters.format_xml(xml_str)
    expected_xml = "<root>\n  <child>text</child>\n</root>"
    assert formatted_xml == expected_xml


def test_format_xml_with_element():
    xml_elem = ET.Element("root")
    child = ET.SubElement(xml_elem, "child")
    child.text = "text"
    formatted_xml = htmlfilters.format_xml(xml_elem)
    expected_xml = "<root>\n  <child>text</child>\n</root>"
    assert formatted_xml == expected_xml


def test_format_xml_with_indent():
    xml_str = "<root><child>text</child></root>"
    formatted_xml = htmlfilters.format_xml(xml_str, indent=4)
    expected_xml = "<root>\n    <child>text</child>\n</root>"
    assert formatted_xml == expected_xml


def test_format_xml_with_level():
    xml_str = "<root><child>text</child></root>"
    formatted_xml = htmlfilters.format_xml(xml_str, level=1)
    # IMO missing indent at beginning this is unexpected, but thats what ET returns..
    expected_xml = "<root>\n    <child>text</child>\n  </root>"
    assert formatted_xml == expected_xml


def test_get_relative_url_empty():
    for url in ["", ".", "/."]:
        for other in ["", ".", "/", "/."]:
            assert htmlfilters.relative_url_mkdocs(url, other) == "."
    assert htmlfilters.relative_url_mkdocs("/", "") == "./"
    assert htmlfilters.relative_url_mkdocs("/", "/") == "./"
    assert htmlfilters.relative_url_mkdocs("/", ".") == "./"
    assert htmlfilters.relative_url_mkdocs("/", "/.") == "./"


if __name__ == "__main__":
    pytest.main([__file__])
